# -*- coding: utf-8 -*-
"""Convience module. 'from bs_ds.imports import *' will pre-load pd,np,plt,mpl,sns"""

from os import link


def global_imports(modulename,shortname = None, asfunction = False,check_vers=True):
        """from stackoverflow: https://stackoverflow.com/questions/11990556/how-to-make-global-imports-from-a-function,
        https://stackoverflow.com/a/46878490"""
        from importlib import import_module

        if shortname is None:
            shortname = modulename

        if asfunction is False:
            globals()[shortname] = import_module(modulename) #__import__(modulename)
        else:
            globals()[shortname] = eval(modulename + "." + shortname)
            
        if check_vers:
            return globals()[shortname].__version__

def clear():
    """Helper function to clear notebook display"""
    import IPython.display as dp
    return dp.clear_output()

def import_packages(import_list_of_tuples = None,  display_table=True, 
                    check_versions=True, link_text=None,
                    check_packages = ['matplotlib','seaborn','pandas','numpy','sklearn','dojo_ds'] ): #append_to_default_list=True, imports_have_description = True):
    """Uses the exec function to load in a list of tuples with:
    [('module','md','example generic tuple item')] formatting.
    >> Default imports_list:
    [('pandas',     'pd',   'High performance data structures and tools'),
    ('numpy',       'np',   'scientific computing with Python'),
    ('matplotlib',  'mpl',  "Matplotlib's base OOP module with formatting artists"),
    ('matplotlib.pyplot',   'plt',  "Matplotlib's matlab-like plotting module"),
    ('seaborn',     'sns',  "High-level data visualization library based on matplotlib"),
    ('IPython.display','dp','Display modules with helpful display and clearing commands.')
    ('dojo_ds','fs','Custom data science bootcamp student package')]
    """


    # import_list=[]
    from IPython.display import display
    import pandas as pd
    # if using default import list, create it:
    if (import_list_of_tuples is None): #or (append_to_default_list is True):
        import_list = [('Package','Handle','Documentation'),
                       ('pandas','pd', "https://pandas.pydata.org/docs/"),#'High performance data structures and tools'),
                       ('dojo_ds','ds',"https://fs-ds.readthedocs.io/en/latest/"),#'Custom data science bootcamp student package'), # 
                       ('numpy','np',"https://numpy.org/doc/stable/reference/"), # 'scientific computing with Python'), #
                       ('matplotlib','mpl','https://matplotlib.org/stable/api/index.html'),#"Matplotlib's base OOP module with formatting artists"), #
                       ('matplotlib.pyplot','plt',"https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.html#module-matplotlib.pyplot"),#"Matplotlib's matlab-like plotting module"), #
                       ('seaborn','sns',"https://seaborn.pydata.org/api.html"),#"High-level data visualization library based on matplotlib"), #
                    #    ('IPython.display','dp',"https://ipython.readthedocs.io/en/stable/api/generated/IPython.display.html")]#'Display modules with helpful display and clearing commands.')]#,
        ]
        # ('cufflinks','cf','Adds df.iplot() interactive Plotly figs. To use, run >> cf.go_offline()')]

    # if using own list, rename to 'import_list'
    else:
        import_list = import_list_of_tuples


    # Use exec command to create global handle variables and then load in package as that handle
    for package,handle,_ in import_list[1:]:
        # old way: # exec(f'import {package} as {handle}')
        # global_imports(package,handle)
         
        global_imports(package,handle,check_vers=False)


    ## Make dataframe of imports
    # create and return styled dataframe
    # columns=['Package','Handle','Documentation']
    df_imports= pd.DataFrame(import_list[1:], columns=import_list[0])
    
    ## Reorder Columns
    # df_imports = df_imports[['Handle','Package','Documentation']]
    
    
  
    
    
    ## make dataframe of versions
    if check_versions:
        pkg_vers_df = check_package_versions(packages=check_packages,
                                             show_only=False,fpath=False)
        
        ## Adding Imported column to track non-imported versions
        df_imports['Imported'] = 'Y'
        df_imports = pd.merge(df_imports, pkg_vers_df,on='Package',how='outer')
        
        df_imports = df_imports[['Package','Handle','Version','Documentation','Imported',]]
        # df_imports.insert(1,'',' as ')
        # df_imports = df_imports.rename({'Package':'import','Handle':'as'},axis=1)
        df_imports['Imported'] = df_imports['Imported'].fillna('N')
        df_imports = df_imports.fillna('')
    

        # display(pkg_vers_df.style.hide_index().set_caption('Package Version Report')) 

    # Display summary dataframe
    if display_table==True:            
        ## Create Columns Names List

        # create and return styled dataframe
        # columns=['Package','Handle','Documentation']
        # df_imported= pd.DataFrame(import_list, columns=columns)
        # # df_imported=pd.concat([df_imported['Handle'],df_imported[['Package','Documentation']]],axis=1)
        
        # # df_imports = pd.merge(df_imported,pkg_vers_df,on='Package')
        # # df_imports = df_imports[['Handle','Package','Documentation','Version']]
        # df_imports = df_imported[['Handle','Package','Documentation']]
        #.sort_values('Package').
        import dojo_ds as ds
        try:
            print(f"dojo_ds v{ds.__version__} loaded.")#  Read the docs: https://fs-ds.readthedocs.io/en/latest/ ")
        except:
            pass
        try:
            dfs = df_imports.style.hide_index().set_caption('Loaded Packages & Info')
        except:
            dfs = df_imports.style.hide().set_caption('Loaded Packages & Info')
        
        ## Determine if links will have display text
        if link_text is None:
            kwargs = {'Documentation':clickable}

        else:
            kwargs = {'Documentation':lambda x:  clickable(x,link_text)}

        ## apply kwargs above and set additional properties
        display(dfs.format(kwargs).set_properties(**{'text-align':'left'}).\
                                set_properties(subset=['Imported','Handle'],**{'text-align':'center'}) )
            # display(dfs.format({'Documentation':lambda x:  clickable(x,link_text)}))
        # return df_imports

    # or just print statement
    else:
        print('Modules successfully loaded.')
        

def clickable(path,label=None):
    """Adapted from: https://www.geeksforgeeks.org/how-to-create-a-table-with-clickable-hyperlink-to-a-local-file-in-pandas/"""
    # returns the final component of a url
    # f_url = os.path.basename(path)
    if label is None:
    # convert the url into link
        return '<a href="{}">{}</a>'.format(path, path)
    else: 
        return '<a href="{}">{}</a>'.format(path, label)  



def check_package_versions(packages = ['matplotlib','seaborn','pandas','numpy','sklearn','dojo_ds'],
                           fpath=False, show_only=True):
    """Imports packages and saves the name and version number to a dataframe"""
    import pandas as pd
    import inspect
    version_list = [['Package','Version']]
    
    ## Remove submodules from version check (wont have version #)
    for package in packages:
        if '.' not in package:
            try:
                ## use global imports and retrieve version #
                vers = global_imports(package,None,check_vers=True)
            except:
                vers = '!'
            version_list.append([package,vers])

    # Convert to df
    pkg_vers_df = pd.DataFrame(version_list[1:],columns=version_list[0])
    
    ## If get_fpath
    if fpath==True:
        pkg_vers_df['File'] = pkg_vers_df['Package'].map(lambda x: inspect.getsourcefile(globals()[x]))
        # for package in packages:
    
    if show_only==True: 
        if fpath==True:   
            dfs = pkg_vers_df.style.set_properties(subset='File',
                                                **{'width':"600px","text-align":'center'})
        else:
            dfs = pkg_vers_df.style
        display(dfs.set_caption('Package Versions'))
    
        
    else:
        return pkg_vers_df



try:
    from IPython.display import clear_output
    clear_output()
except:
    pass
finally:
    # fs = None
    import_packages()
    
# try:
#     import cufflinks as cf 
#     cf.go_offline()
#     print('[i] Pandas .iplot() method activated.')
# except:
#     pass

    


import statsmodels.tsa.api as tsa
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import statsmodels.tsa.api as tsa
import numpy as np
import statsmodels.tsa.api as tsa
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score, mean_absolute_percentage_error

def get_adfuller_results(ts, alpha=.05, label='adfuller', **kwargs):
    """
    Uses statsmodels' adfuller function to test a univariate time series for stationarity.
    Null hypothesis: The time series is NOT stationary. (It "has a unit root".)
    Interpretation: A p-value less than alpha (.05) means the ts IS stationary.
    (We reject the null hypothesis that it is not stationary.)

    Parameters:
    - ts (array-like): The time series data.
    - alpha (float): The significance level used for interpreting p-value. Default is 0.05.
    - label (str): The label for the output DataFrame. Default is 'adfuller'.
    - **kwargs: Additional keyword arguments to be passed to statsmodels' adfuller function.

    Returns:
    - results (DataFrame): DataFrame with the following columns/results:
        - "Test Statistic": The adfuller test statistic.
        - "# of Lags Used": The number of lags used in the calculation.
        - "# of Observations": The number of observations used.
        - "p-value": The p-value for the hypothesis test.
        - "alpha": The significance level used for interpreting p-value.
        - "sig/stationary?": A simplified interpretation of the p-value.

    ADFULLER DOCUMENTATION:
    For the full adfuller documentation, see:
    https://www.statsmodels.org/dev/generated/statsmodels.tsa.stattools.adfuller.html
    """
    (test_stat, pval, nlags, nobs, crit_vals_d, icbest) = tsa.adfuller(ts, **kwargs)
    adfuller_results = {
        'Test Statistic': test_stat,
        '# of Lags Used': nlags,
        '# of Observations': nobs,
        'p-value': round(pval, 6),
        'alpha': alpha,
        'sig/stationary?': pval < alpha
    }
    return pd.DataFrame(adfuller_results, index=[label])


def get_sig_lags(ts, type='ACF', nlags=None, alpha=0.5):
    """
    Calculates the significant lags for the autocorrelation function (ACF) or partial autocorrelation function (PACF) of a time series.

    Parameters:
    - ts (array-like): The time series data.
    - type (str): The type of lag to calculate. Must be either 'ACF' or 'PACF'. Default is 'ACF'.
    - nlags (int): The number of lags to calculate. Default is None.
    - alpha (float): The significance level for determining significant lags. Default is 0.5.

    Returns:
    - sig_lags (array-like): The significant lags.

    Raises:
    - Exception: If the type is not 'ACF' or 'PACF'.
    """
    if type == 'ACF':
        corr_values, conf_int = tsa.stattools.acf(ts, alpha=alpha, nlags=nlags)
    elif type == 'PACF':
        corr_values, conf_int = tsa.stattools.pacf(ts, alpha=alpha, nlags=nlags)
    else:
        raise Exception("type must be either 'ACF' or 'PACF'")

    lags = range(len(corr_values))
    corr_df = pd.DataFrame({
        type: corr_values,
        'Lags': lags,
        'lower ci': conf_int[:, 0] - corr_values,
        'upper ci': conf_int[:, 1] - corr_values
    })
    corr_df = corr_df.set_index("Lags")

    filter_sig_lags = (corr_df[type] < corr_df['lower ci']) | (corr_df[type] > corr_df['upper ci'])
    sig_lags = corr_df.index[filter_sig_lags]
    sig_lags = sig_lags[sig_lags != 0]

    return sig_lags


def plot_acf_pacf(ts, nlags=40, figsize=(10, 5), annotate_sig=False, alpha=.05, acf_kws={}, pacf_kws={}, annotate_seas=False, m=None, seas_color='black'):
    """
    Plots the autocorrelation function (ACF) and partial autocorrelation function (PACF) of a time series.

    Parameters:
    - ts (array-like): The time series data.
    - nlags (int): The number of lags to include in the plot. Default is 40.
    - figsize (tuple): The size of the figure. Default is (10, 5).
    - annotate_sig (bool): Whether to annotate significant lags. Default is False.
    - alpha (float): The significance level for determining significant lags. Default is 0.05.
    - acf_kws (dict): Additional keyword arguments to be passed to statsmodels' plot_acf function.
    - pacf_kws (dict): Additional keyword arguments to be passed to statsmodels' plot_pacf function.
    - annotate_seas (bool): Whether to annotate seasonal lags. Default is False.
    - m (int): The number of observations per season. Required if annotate_seas is True.
    - seas_color (str): The color of the seasonal lines. Default is 'black'.

    Returns:
    - fig (Figure): The matplotlib Figure object.
    """
    fig, axes = plt.subplots(nrows=2, figsize=figsize)
    sig_vline_kwargs = dict(ls=':', lw=1, zorder=0, color='red')

    tsa.graphics.plot_acf(ts, ax=axes[0], lags=nlags, **acf_kws)

    if annotate_sig:
        sig_acf_lags = get_sig_lags(ts, nlags=nlags, alpha=alpha, type='ACF')
        for lag in sig_acf_lags:
            axes[0].axvline(lag, label='sig', **sig_vline_kwargs)

    tsa.graphics.plot_pacf(ts, ax=axes[1], lags=nlags, **pacf_kws)

    if annotate_sig:
        sig_pacf_lags = get_sig_lags(ts, nlags=nlags, alpha=alpha, type='PACF')
        for lag in sig_pacf_lags:
            axes[1].axvline(lag, label='sig', **sig_vline_kwargs)

    if annotate_seas:
        if m is None:
            raise Exception("Must define value of m if annotate_seas=True.")

        n_seasons = nlags // m
        seas_vline_kwargs = dict(ls='--', lw=1, alpha=.7, color=seas_color, zorder=-1)

        for i in range(1, n_seasons + 1):
            axes[0].axvline(m * i, **seas_vline_kwargs, label="season")
            axes[1].axvline(m * i, **seas_vline_kwargs, label="season")

    fig.tight_layout()
    return fig


def plot_forecast(ts_train, ts_test, forecast_df, n_train_lags=None, 
                  figsize=(10,4), title='Comparing Forecast vs. True Data'):
    """
    Plots the training data, test data, and forecast with confidence intervals.

    Parameters:
    ts_train (pandas.Series): Time series data for training.
    ts_test (pandas.Series): Time series data for testing.
    forecast_df (pandas.DataFrame): DataFrame containing forecasted values and confidence intervals.
    n_train_lags (int, optional): Number of training lags to plot. If not specified, all training data will be plotted.
    figsize (tuple, optional): Figure size. Default is (10, 4).
    title (str, optional): Title of the plot. Default is 'Comparing Forecast vs. True Data'.

    Returns:
    fig (matplotlib.figure.Figure): The generated figure.
    ax (matplotlib.axes.Axes): The axes of the generated figure.
    """
    ### PLot training data, and forecast (with upper/,lower ci)
    fig, ax = plt.subplots(figsize=figsize)

    # setting the number of train lags to plot if not specified
    if n_train_lags==None:
        n_train_lags = len(ts_train)
            
    # Plotting Training  and test data
    ts_train.iloc[-n_train_lags:].plot(ax=ax, label="train")
    ts_test.plot(label="test", ax=ax)

    # Plot forecast
    forecast_df['mean'].plot(ax=ax, color='green', label="forecast")

    # Add the shaded confidence interval
    ax.fill_between(forecast_df.index, 
                    forecast_df['mean_ci_lower'],
                   forecast_df['mean_ci_upper'],
                   color='green', alpha=0.3,  lw=2)

    # set the title and add legend
    ax.set_title(title)
    ax.legend();
    
    return fig, ax


def thiels_U(ts_true, ts_pred):
    """Calculate's Thiel's U metric for forecasting accuracy.
    Accepts true values and predicted values.
    Original Formula Source: https://docs.oracle.com/cd/E57185_01/CBREG/ch06s02s03s04.html
    Adapted Function from Source: https://github.com/jirvingphd/predicting-the-SP500-using-trumps-tweets_capstone-project/blob/cf11f6ed88721433d2c00cb1f8486206ab179cc0/bsds/my_keras_functions.py#L735
    Returns: 
        U (float)
        
    Thiel's U Value Interpretation:
    - <1  = Forecasting is better than guessing 
    - 1   = Forecasting is about as good as guessing
    - >1  = Forecasting is worse than guessing 
    """
    import numpy as np
    # sum_list = []
    num_list=[]
    denom_list=[]
    
    for t in range(len(ts_true)-1):
        
        num_exp = (ts_pred[t+1] - ts_true[t+1])/ts_true[t]
        num_list.append([num_exp**2])
        
        denom_exp = (ts_true[t+1] - ts_true[t])/ts_true[t]
        denom_list.append([denom_exp**2])
        
    U = np.sqrt( np.sum(num_list) / np.sum(denom_list))
    return U


def calc_thiels_U(*args, **kwargs):
    return thiels_U(*args, **kwargs)


def regression_metrics_ts(ts_true, ts_pred, label="", verbose=True, output_dict=False,
                          thiels_U=False):
    """
    Calculates regression metrics for comparing true and predicted time series data.

    Parameters:
    - ts_true (array-like): The true time series data.
    - ts_pred (array-like): The predicted time series data.
    - label (str): The label for the metrics. Default is an empty string.
    - verbose (bool): Whether to print the metrics. Default is True.
    - output_dict (bool): Whether to return the metrics as a dictionary. Default is False.
    - thiels_U (bool): Whether to calculate Thiel's U metric. Default is False.
                        - See docstring for thiels_U function for interpretation.

    Returns:
    - metrics (dict): The regression metrics as a dictionary. Only returned if output_dict is True.
    """
    mae = mean_absolute_error(ts_true, ts_pred)
    mse = mean_squared_error(ts_true, ts_pred)
    rmse = mean_squared_error(ts_true, ts_pred, squared=False)
    r_squared = r2_score(ts_true, ts_pred)
    mae_perc = mean_absolute_percentage_error(ts_true, ts_pred) * 100
    if thiels_U==True:
        U = calc_thiels_U(ts_true, ts_pred)
    if verbose:
        header = "---" * 20
        print(header, f"Regression Metrics: {label}", header, sep="\n")
        print(f"- MAE = {mae:,.3f}")
        print(f"- MSE = {mse:,.3f}")
        print(f"- RMSE = {rmse:,.3f}")
        print(f"- R^2 = {r_squared:,.3f}")
        print(f"- MAPE = {mae_perc:,.2f}%")
        
        if thiels_U:
            print(f"- Thiel's U = {U:,.2f}")
        print(header)
            
    if output_dict:
        metrics = {
            "Label": label,
            "MAE": mae,
            "MSE": mse,
            "RMSE": rmse,
            "R^2": r_squared,
            "MAPE(%)": mae_perc,
        }
        if thiels_U:
            metrics['Thiel\'s U'] = U
        return metrics





def plot_acf(ts, nlags=None, alpha=0.05, figsize=(12,4), annotate_sig=False): 
    """
    Plots the autocorrelation function (ACF) for a given time series.

    Parameters:
    - ts: The time series data.
    - nlags: The number of lags to include in the plot. If None, all lags are included.
    - alpha: The significance level for determining significant lags.
    - figsize: The size of the figure (width, height) in inches.
    - annotate_sig: Whether to annotate significant lags with red dashed lines.

    Returns:
    - None
    """
    fig, ax = plt.subplots(figsize=figsize) 
    tsa.graphics.plot_acf(ts, ax=ax, lags=nlags, alpha=alpha)

    if annotate_sig == True:
        sig_lags = get_sig_lags(ts,nlags=nlags,alpha=alpha, type='ACF')
        
        ## ADDING ANNOTATING SIG LAGS
        for lag in sig_lags:
            ax.axvline(lag, ls='--', lw=1, zorder=0, color='red')
            
            

def plot_pacf(ts, nlags=None, alpha=0.05, figsize=(12,4), annotate_sig=False): 
    """
    Plots the partial autocorrelation function (PACF) for a given time series.

    Parameters:
    - ts: The time series data.
    - nlags: The number of lags to include in the plot. If None, all lags will be included.
    - alpha: The significance level for determining significant lags.
    - figsize: The size of the figure (width, height) in inches.
    - annotate_sig: Whether to annotate significant lags with red dashed lines.

    Returns:
    - None
    """
    fig, ax = plt.subplots(figsize=figsize) 
    tsa.graphics.plot_pacf(ts, ax=ax, lags=nlags, alpha=alpha)

    if annotate_sig == True:
        sig_lags = get_sig_lags(ts,nlags=nlags,alpha=alpha, type='PACF')
        
        ## ADDING ANNOTATING SIG LAGS
        for lag in sig_lags:
            ax.axvline(lag, ls='--', lw=1, zorder=0, color='red')


def display_ts_workflow(return_text=False):
    """
    Displays the Time Series Modeling Workflow - (S)ARIMA markdown document.

    Parameters:
    - return_text (bool): If True, returns the text content of the markdown document as a string. Default is False.

    Returns:
    - str: The text content of the markdown document, if return_text is True.

    """
    import requests
    try:
        from IPython.display import Markdown, display
        display_func = lambda x: display(Markdown(x))
    except:
        display_func = print
    url = "https://raw.githubusercontent.com/coding-dojo-data-science/dojo_ds/main/assets/Time%20Series%20Modeling%20Workflow%20-%20(S)ARIMA.md"
    resp = requests.get(url)
    display_func(resp.text)

    if return_text:
        return resp.text









# def reference_interpreting_acf_pacf(seasonal=True, return_string = False):
#     table = """**Determining (S)ARIMA Orders from ACF/PACF**:\n
# |                  | AR(p)                | MA(q)               | ARMA(p,q) |
# |:----- | :---------------: | :---------------:|:---------------:|
# | **ACF**              | Gradually decreases  | Dramatic drop after lag \(q\) | Gradually decreases |
# | **PACF**             | Dramatic drop after lag \(p\) | Gradually decreases      | Gradually decreases  |
# | **ARIMA Order** (p,d,q) | (p,d,0) | (0,d,q) | Start with (1,d,1) & Iterate|

# ___

# **If seasonality is present**

# | ***If seasonal*** | S-AR(P)                | S-MA(Q)                | SARIMA(P,Q)            |
# |:-----------------| :---------------: | :---------------:|:---------------:|  
# | **ACF** (seasonal lags)†         | Gradually decreases   | Dramatic drop after lag \(Q\)| Gradually decreases | 
# | **PACF** (seasonal lags)†       |  Dramatic drop after lag \(P\) | Gradually decreases    | Gradually decreases |
# | **Seasonal Order** (P,D,Q)  | (P,D,0)      | (0,D,Q)       | Start with (1,D,1) & Iterate  | 

# - † seasonal lags = lags that are a multiple of the season length (m). E.g., If daily, m=7, check lags 7,14,21,etc.

# """
#     if return_string == True:
#         return table
#     else: 
#         from IPython.display import Markdown, display
#         display(Markdown(table))


def reference_interpreting_acf_pacf(seasonal=True, return_string=False):
    """
    Generates a table for determining (S)ARIMA orders from ACF/PACF.

    Args:
        seasonal (bool, optional): Indicates if seasonality is present. Defaults to True.
        return_string (bool, optional): Indicates if the table should be returned as a string. 
            If False, the table is displayed using IPython's Markdown. Defaults to False.

    Returns:
        str or None: The table as a string if return_string is True, otherwise None.
    """
    table = """#### **Determining (S)ARIMA Orders from ACF/PACF**:\n
|                  | AR(p)                | MA(q)               | ARMA(p,q) |
|:----- | :---------------: | :---------------:|:---------------:|
| **ACF**              | Gradually decreases  | Dramatic drop after lag \(q\) | Gradually decreases |
| **PACF**             | Dramatic drop after lag \(p\) | Gradually decreases      | Gradually decreases  |
| **ARIMA Order (p,d,q)** | **(p,d,0)** | **(0,d,q)** | **Start with (1,d,1) & Iterate**|


___


#### **Determining Seasonal Orders** (only if seasonality is present):

| ***If seasonal*** | S-AR(P)                | S-MA(Q)                | SARIMA(P,Q)            |
|:-----------------| :---------------: | :---------------:|:---------------:|  
| **ACF** (seasonal lags)†         | Gradually decreases   | Dramatic drop after lag \(Q\)| Gradually decreases | 
| **PACF** (seasonal lags)†       |  Dramatic drop after lag \(P\) | Gradually decreases    | Gradually decreases |
| **Seasonal Order (P,D,Q)**  | **(P,D,0)**      | **(0,D,Q)**       | **Start with (1,D,1) & Iterate**  | 

- † seasonal lags = lags that are a multiple of the season length (m). E.g., If daily, m=7, check lags 7,14,21,etc.

"""
    if seasonal==False:
        table = table.split('___')[0]
        table = table.replace("(S)ARIMA",'ARIMA')
   
    
             
    if return_string:
        return table
    else: 
        try:
            from IPython.display import Markdown, display
            display_func = lambda x: display(Markdown(x))
        except:
            display_func = print
            
        display_func(table)



def make_best_arima(auto_model, ts_train, exog=None, fit_kws={}):
    """
    Fits a final ARIMA model using the best parameters obtained from an auto_model and evaluates its performance.

    Parameters:
    auto_model (AutoARIMA): The AutoARIMA model object that contains the best parameters.
    ts_train (array-like): The time series data used for training the ARIMA model.
    exog (array-like, optional): Exogenous variables to be included in the model. Default is None.
    fit_kws (dict, optional): Additional keyword arguments to be passed to the `fit` method of the SARIMAX model. Default is an empty dictionary.

    Returns:
    SARIMAX: The fitted SARIMAX model with the best parameters.

    """
    best_model = tsa.SARIMAX(
        ts_train,
        exog=exog,
        order=auto_model.order,
        seasonal_order=auto_model.seasonal_order,
        sarimax_kwargs=auto_model.sarimax_kwargs,
    ).fit(disp=False, **fit_kws)
    return best_model

    
            
        
def evaluate_ts_model(model, ts_train, ts_test, exog_train=None, exog_test=None,
                      return_scores=False, show_summary=True,
                      n_train_lags=None, figsize=(9,3),
                      title='Comparing Forecast vs. True Data',
                     plot_diagnostics=True):
    """
    Evaluates a time series model by generating forecasts and comparing them to the true data.

    Parameters:
    - model: The time series model to be evaluated (Either SARIMAX or AutoARIMA object)
    - ts_train: The training time series data.
    - ts_test: The testing time series data.
    - exog_train: The exogenous variables for the training data (optional).
    - exog_test: The exogenous variables for the testing data (optional).
    - return_scores: Whether to return the model and regression metrics (optional, default=False).
    - show_summary: Whether to display the model summary and diagnostics plots (optional, default=True).
    - n_train_lags: The number of lagged values to include in the training data visualization (optional).
    - figsize: The size of the forecast plot (optional, default=(9,3)).
    - title: The title of the forecast plot (optional, default='Comparing Forecast vs. True Data').

    Returns:
    - If return_scores=True, returns the model and regression metrics.
    - If return_scores=False, returns the model.

    """
    import warnings

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        ## GET FORECAST             
        # Check if auto-arima, if so, extract sarima model
        if hasattr(model, "arima_res_"):
            print(f"- Fitting a new ARIMA using the params from the auto_arima...")
            model = make_best_arima(model, ts_train, exog=exog_train)
            

                
    
        # Get forecast         
        forecast = model.get_forecast(exog=exog_test, steps=len(ts_test))
        forecast_df = forecast.summary_frame()
                        
        # Get and display the regression metrics BEFORE showing plot
        reg_res = regression_metrics_ts(ts_test, forecast_df['mean'], 
                                            output_dict=True, thiels_U=True)
        
        if show_summary==True:
            display(model.summary())
            
        if plot_diagnostics==True:
            model.plot_diagnostics(figsize=(8,4))
            plt.tight_layout()
            plt.show()
        # Visualize forecast
        plot_forecast(ts_train, ts_test, forecast_df, 
                      n_train_lags=n_train_lags, figsize=figsize,
                     title=title)
        plt.show()
    
        
    
        if return_scores:
            return model, reg_res
        else:
            return model



"""Filesize analysis functions.
Created with ChatGPT.

# Example usage

# Example usage
>> directory_path = '/path/to/your/directory'
>> df = list_files_in_directory(directory_path, size_threshold=100, recursive=True, return_df=True, details=True, unit='MB', sort=True, only_large=True)
>> display(df)

"""
import os
import re
import pandas as pd
from datetime import datetime
import stat

def get_file_size(size_in_bytes, unit=None):
    if unit is None:
        if size_in_bytes < 1024:
            return f"{size_in_bytes} B", size_in_bytes / (1024**2)
        elif size_in_bytes < 1024**2:
            return f"{size_in_bytes / 1024:.2f} KB", size_in_bytes / (1024**2)
        elif size_in_bytes < 1024**3:
            return f"{size_in_bytes / (1024**2):.2f} MB", size_in_bytes / (1024**2)
        else:
            return f"{size_in_bytes / (1024**3):.2f} GB", size_in_bytes / (1024**2)
    else: 
        unit = unit.upper()
        if unit == 'B':
            return f"{size_in_bytes} B", size_in_bytes / (1024**2)
        elif unit == 'KB':
            return f"{size_in_bytes / 1024:.2f} KB", size_in_bytes / (1024**2)
        elif unit == 'MB':
            return f"{size_in_bytes / (1024**2):.2f} MB", size_in_bytes / (1024**2)
        elif unit == 'GB':
            return f"{size_in_bytes / (1024**3):.2f} GB", size_in_bytes / (1024**2)

def get_permissions_text(mode):
    """Generate a string representing file permissions in Unix format."""
    is_dir = 'd' if stat.S_ISDIR(mode) else '-'
    perm_str = is_dir + \
               ''.join([(('r', '-')[bool(mode & perm)] +
                         ('w', '-')[bool(mode & perm)] +
                         ('x', '-')[bool(mode & perm)]) for perm in [stat.S_IRUSR, stat.S_IWUSR, stat.S_IXUSR,
                                                                    stat.S_IRGRP, stat.S_IWGRP, stat.S_IXGRP,
                                                                    stat.S_IROTH, stat.S_IWOTH, stat.S_IXOTH]])
    return perm_str

def get_filetype_emoji(file_type):
    """Return an emoji representing the file type."""
    file_type_emojis = {
        '.txt': '📄',
        '.pdf': '📄',
        '.doc': '📄',
        '.docx': '📄',
        '.xls': '📊',
        '.xlsx': '📊',
        '.csv': '📊',
        '.py': '🐍',
        '.jpg': '🖼️',
        '.jpeg': '🖼️',
        '.png': '🖼️',
        '.gif': '🖼️',
        '.zip': '🗜️',
        '.tar': '🗜️',
        '.gz': '🗜️',
        '.mp3': '🎵',
        '.wav': '🎵',
        '.mp4': '🎥',
        '.mkv': '🎥',
        '.html': '🌐',
        '.css': '🎨',
        '.js': '💻',
        '.json': '🔧',
        '.xml': '🔧',
        '.yaml': '🔧',
        '.yml': '🔧',
        '.joblib': '🔧',
        '.pb': '🧠',
        '.h5': '🧠',
        '.tfrecord': '🧠',
        '.index': '🧠',
        '.data': '🧠',
        '.ckpt': '🧠',
        '.tflite': '🧠',
        '.tf record': '🧠',
        '.pth': '🧠',
        '.pt': '🧠',
        '.onnx': '🧠',
        '.pkl': '🧠',
        '.pickle': '🧠',
        '.hdf5': '🧠',
        '.npy': '🧠',
        '.safetensor': '🧠',
    }
    # Check for exact match in predefined extensions
    if file_type in file_type_emojis:
        return file_type_emojis[file_type]
    
    
    return '📁'

def is_path_compatible(file_path):
    """Check if the file path is compatible with Windows, macOS, and Linux."""
    windows_invalid_chars = r'[<>:"/\\|?*\x00-\x1F]'
    mac_invalid_chars = r'[:]'
    linux_invalid_chars = r'[\x00]'
    
    windows_invalid = bool(re.search(windows_invalid_chars, file_path))
    mac_invalid = bool(re.search(mac_invalid_chars, file_path))
    linux_invalid = bool(re.search(linux_invalid_chars, file_path))
    
    return not (windows_invalid or mac_invalid or linux_invalid)

def list_files_in_directory(directory, size_threshold=100, recursive=True, return_df=True, details=False, unit='MB', sort=False, only_large=False, include_dir_details=False, check_compatibility=False, only_incompatible=False):
    """
    List files in a directory and its subdirectories.

    Args:
        directory (str): The path to the directory.
        size_threshold (int, optional): The size threshold in the specified units. Files larger than this threshold will be considered as "large". Defaults to 100.
        recursive (bool, optional): Whether to list files recursively in subdirectories. Defaults to False.
        return_df (bool, optional): Whether to return a DataFrame. If False, prints the results. Defaults to True.
        details (bool, optional): Whether to include detailed information (owner, permissions). Defaults to False.
        unit (str, optional): The unit for displaying file sizes. Defaults to 'MB'.
        sort (bool, optional): Whether to sort the results from largest to smallest. Defaults to False.
        only_large (bool, optional): Whether to include only files larger than the size threshold. Defaults to False.
        include_dir_details (bool, optional): Whether to include total size and number of files for directories. Defaults to False.
        check_compatibility (bool, optional): Whether to check if file paths are compatible with all OS. Defaults to False.
        only_incompatible (bool, optional): Whether to include only files with incompatible paths. Defaults to False.

    Returns:
        pandas.DataFrame: A DataFrame containing information about the files in the directory if return_df is True.

    Example usage:
        directory_path = '/path/to/your/directory'
        df = list_files_in_directory(directory_path, size_threshold=100, recursive=True, return_df=True, details=True, unit='MB', sort=True, only_large=True, include_dir_details=True, check_compatibility=True, only_incompatible=False)
        print(df)
    """
    file_list = []

    # Convert the size threshold to bytes based on the specified unit
    if unit == 'B':
        size_threshold_bytes = size_threshold
    elif unit == 'KB':
        size_threshold_bytes = size_threshold * 1024
    elif unit == 'MB':
        size_threshold_bytes = size_threshold * 1024 * 1024
    elif unit == 'GB':
        size_threshold_bytes = size_threshold * 1024 * 1024 * 1024
    else:
        raise ValueError("Invalid unit. Please use 'B', 'KB', 'MB', or 'GB'.")
    
    # Traverse the directory
    for root, dirs, files in os.walk(directory):
        # Add directories to the list if not recursive
        if not recursive or include_dir_details:
            for dir_name in dirs:
                dir_path = os.path.join(root, dir_name)
                dir_stats = os.stat(dir_path)
                
                # Get directory creation and modification dates
                creation_date = datetime.fromtimestamp(dir_stats.st_ctime).strftime('%Y-%m-%d %H:%M:%S')
                modification_date = datetime.fromtimestamp(dir_stats.st_mtime).strftime('%Y-%m-%d %H:%M:%S')
                
                # Initialize directory size and file count
                dir_size = 0
                file_count = 0
                
                # Calculate total size and number of files if requested
                if include_dir_details:
                    for dir_root, _, dir_files in os.walk(dir_path):
                        for f in dir_files:
                            file_count += 1
                            f_path = os.path.join(dir_root, f)
                            dir_size += os.path.getsize(f_path)
                
                dir_size_str, _ = get_file_size(dir_size, unit=unit)
                
                # Check path compatibility if requested
                path_compatible = True
                if check_compatibility:
                    path_compatible = is_path_compatible(dir_path)
                    if only_incompatible and path_compatible:
                        continue
                
                # Create a dictionary of directory details
                dir_info = {
                    'Name': dir_name,
                    'Absolute Path': dir_path,
                    'Size': f"{file_count} Files",
                    'Size (MB)': dir_size / (1024**2),
                    'Creation Date': creation_date,
                    'Modification Date': modification_date,
                    'File Type': 'directory',
                    'File Type Icon': '🗂️',
                    'Is Large File': False,
                    # 'Path Compatible': path_compatible
                }
                if check_compatibility:
                    dir_info['Path Compatible'] = path_compatible
                
                if return_df:
                    file_list.append(dir_info)
                else:
                    # Print directory details in a professional format
                    for key, value in dir_info.items():
                        if key not in {'File Type Icon', 'Name', 'Is Large File'}:
                            print(f"{key:25}: {value}")
                    print("-" * 80)
        
        for file in files:
            file_path = os.path.join(root, file)
            file_stats = os.stat(file_path)
            
            # Gather file details
            file_size = file_stats.st_size
            size_str, size_in_mb = get_file_size(file_size, unit=unit)
            is_large_file = file_size > size_threshold_bytes
            
            if only_large and not is_large_file:
                continue
            
            # Get file creation and modification dates
            creation_date = datetime.fromtimestamp(file_stats.st_ctime).strftime('%Y-%m-%d %H:%M:%S')
            modification_date = datetime.fromtimestamp(file_stats.st_mtime).strftime('%Y-%m-%d %H:%M:%S')

            # Determine file type and its corresponding emoji
            file_type = os.path.splitext(file)[-1].lower()
            if re.match(r'\.data-\d{5}-of-\d{5}', file_type):
                file_type = 'tf record'  # Normalize to generic '.data' file type
                
            file_type_icon = get_filetype_emoji(file_type)
            # Check path compatibility if requested
            path_compatible = True
            if check_compatibility:
                path_compatible = is_path_compatible(file_path)
                if only_incompatible and path_compatible:
                    continue

            # Create a dictionary of file details
            file_info = {
                'Name': file,
                'Absolute Path': file_path,
                'Size': size_str,
                'Size (MB)': size_in_mb,
                'Creation Date': creation_date,
                'Modification Date': modification_date,
                'File Type': file_type,
                'File Type Icon': file_type_icon,
                'Is Large File': is_large_file,
            }
            if check_compatibility:
                file_info['Path Compatible'] = path_compatible

            # Include detailed information if requested
            if details:
                file_info['File Owner'] = file_stats.st_uid
                file_info['File Permissions (Octal)'] = oct(file_stats.st_mode & 0o777)
                file_info['File Permissions (Text)'] = get_permissions_text(file_stats.st_mode)

            file_list.append(file_info)
        
        # If not recursive, break after processing the top directory
        if not recursive:
            break
    
    # Sort the list if requested
    if sort:
        file_list.sort(key=lambda x: x['Size (MB)'], reverse=True)
    
    # Return the DataFrame if requested
    if return_df:
        df = pd.DataFrame(file_list)
        # Reorder the columns
        cols = ['Name', 'File Type', 'File Type Icon', 'Size', 'Creation Date', 'Modification Date'] + [col for col in df.columns if col not in ['Name', 'File Type', 'File Type Icon', 'Size', 'Creation Date', 'Modification Date']]
        df = df[cols].copy()
        return df
    else:
        # Print file details in a professional format, excluding the icon, name, and is_large_file
        for file_info in file_list:
            for key, value in file_info.items():
                if key not in {'File Type Icon', 'Name', 'Is Large File'}:
                    print(f"{key:25}: {value}")
            print("-" * 80)



"""
Custom Functions the Data Science Program
__author__ = James Irving, Brenda Hungerford
"""
__author__ = """James Irving"""
__email__ = 'james.irving.phd@gmail.com'
__version__ = '1.1.15'
from . import data_enrichment as data
from . import eda as eda
from . import evaluate as evaluate
from . import insights  as insights
from . import nlp as nlp
from . import time_series as time_series
from . import datasets as datasets
from . import fileinfo as fileinfo

# from . import deployment_functions as deploy

from . import utils as utils


# from . import _eda_functions_plotly as eda_plotly

def show_code(function):
	"""
	Uses the inspect module to retrieve the source code for a function.
	Displays the code as Python-syntax Markdown code.

	Note: Python highlighting may not work correctly on some editors.

	Parameters:
	function (callable): The function for which to display the source code.

	Returns:
	None
	"""
	import inspect
	from IPython.display import display, Markdown

	code = inspect.getsource(function)
	md = "```python" +'\n' + code + "\n" + '```' 
	display(Markdown(md))

